#!/usr/bin/env python
# -*- coding: utf-8 -*-
# raspEasyFire, die RaspberryPi - Feuerwerskzündanlage
# Author, Jens Nachtigal - info@newan.de
# Weiternutzung und Veränderung bitte mit absprache


import pygame
from Queue import PriorityQueue
import csv
import datetime
import time
import os
import smbus
import signal

#Imports for LCD / File liegt im selben ordner (dort Adresse setzen)!
import RPi_I2C_driver

# global constants
FREQ              = 44100  # same as audio CD
BITSIZE           = -16    # unsigned 16 bit
CHANNELS          = 2      # 1 == mono, 2 == stereo
BUFFER            = 1024   # audio buffer size in no. of samples
FRAMERATE         = 30     # how often to check if playback has finished
const_adressDict  = {0: 0x20} # I2C Adresse
const_BankDict     = {1: 0x12, 2: 0x13} #Backnadresse hinzufügen A und B - Bank
const_PinDict     = {7: 0x80, 6:0x40, 5:0x20, 4:0x10,3:0x08, 2:0x04, 1:0x02, 0:0x01, "all":0xff, "off":0x00} # I2C Pin


#FireEvent Klasse
class FireEvent:
  def __init__(self, dt, channel):
    self.dt = dt
    self.channel = channel

  def getTime(self):
    return self.dt

  def getChannel(self):
    return self.channel

#EventManager Klasse
class EventManager:
  def __init__(self, objLcd):
    self.quEVENT = PriorityQueue(maxsize=0)
    self.objLcd = objLcd
    self.objBus = smbus.SMBus(1)
    self.portNumbersCache = dict();
    self.sleepTimer = 0
    self.PinCache = []
    self.countDownTime = 9999

  def setNewEvent(self, dt, channel):
    objEvent = FireEvent(dt, channel)
    self.quEVENT.put((dt, time.time(), objEvent))

  def getNextEvent(self):
    if self.quEVENT.empty():
      self.objLcd.lcd_clear()
      self.objLcd.lcd_display_string("  ..Events..  ", 1)
      self.objLcd.lcd_display_string("  ..finished..  ", 2)
      sleep(3)
    else:
      print "Hole next Event"
      #Holen uns das naechste event von der Queue
      objEvent = self.quEVENT.get()[2]
      self.nextFireDt = objEvent.getTime()
      self.nextFireChannel = objEvent.getChannel()
      self.quEVENT.task_done()

  def TimeEvent(self, dt):
    if(self.nextFireDt <= dt):
      self.fire(self.nextFireChannel)
      self.getNextEvent() # load next Event
  
    #updaten das Display
    if (self.sleepTimer <= 0):
      if (self.sleepTimer < 0 or self.countDownTime != abs(self.nextFireDt - dt) / 1000):
        self.countDownTime = abs(self.nextFireDt - dt) / 1000
        self.objLcd.lcd_clear()

      #Wenn wir unter 5 Sekunden sind Coutdown!      
      self.objLcd.lcd_display_string("  ..waiting..  ", 2)
      self.objLcd.lcd_display_string(" Zündung in: " +str(self.countDownTime)+"  ", 1)
      self.sleepTimer = 0 # Display wird nicht mehr geupdated bis wir nächsten Event haben
    else:
      # warten das die anzeige sich aktualisiert
      self.sleepTimer = self.sleepTimer - 0.03
   
  def fire(self, channel):
    #Userinfos
    self.objLcd.lcd_clear()
    self.objLcd.lcd_display_string("  Fire Channel  ", 1)
    self.objLcd.lcd_display_string("  **** " + str(channel) + " ****  ", 2)
    self.sleepTimer = 5 #Todoooo, kein sleep sondenr ein Timer bauen!

    ### the A0/A1/A2 pins can be tied high/low individually, 
    ### so you can have up to 8 on the MCP23017 on the same bus
    if(channel//16 < 8):
      # each MCP23017 hast 16 Pins 0-15, Berechnen unsere Werte
      tmpChipadress = const_adressDict[channel//16];
      if(channel%16 < 8):
        tmpBankadress = const_BankDict[1]
      else:
        tmpBankadress = const_BankDict[2]
      portNumber = channel%16%8
      
      # Lassen den Kanal offen, daher aktivieren wir einen Cache
      isInArray = False
      for obj in self.PinCache:
        if (obj[0] == tmpChipadress and obj[1] == tmpBankadress):
          tmpSendCommand = obj[2] | const_PinDict[portNumber]
          obj[2] = obj[2] | const_PinDict[portNumber] # im Cache ablegen für den nächsten
          isInArray = True

      if (isInArray == False):
        # Chip war nicht im Array
        self.PinCache.append([tmpChipadress, tmpBankadress, const_PinDict[portNumber]])
        tmpSendCommand = const_PinDict[portNumber]
      
      # senden das or verknüfte bit
      print "Senden über den Bus fogendes: write_byte_data("+ hex(tmpChipadress) +","+hex(tmpBankadress)+","+hex(tmpSendCommand)+");" 
      self.objBus.write_byte_data(tmpChipadress, tmpBankadress, tmpSendCommand)     
    else:
      #Channelnumber to high for 8 chips
      raise Exception('Channelnumber is to high')

  def checkMaxChannel(self, nMaxChannel):
    tmpMaxChip = nMaxChannel//16;
    try:
      for x in range(0, tmpMaxChip+1):
        print "Setze Bus für Chip auf 0 für: " + hex(const_adressDict[x])
        self.objBus.write_byte_data(const_adressDict[x], 0x00, 0x00) # Bank A Ausgang und AUS
        self.objBus.write_byte_data(const_adressDict[x], 0x01, 0x00) # Bank B Ausgang und AUS
      return True
    except:
      print "Sind in Exception"
      return False
    

# Hauptklasse / LogikKlasse
class raspEasyFire:

  def __init__(self, objRaspEasyFirePins, csvFile, sTestAudio, sFinalAudio):
    #baue innerreferenzen auf
    self.objRaspEasyFirePins = objRaspEasyFirePins;
    self.objRaspEasyFirePins.setRaspEasyFire(self)

    self.sTestAudio = sTestAudio
    self.sFinalAudio = sFinalAudio
    self.csvFile = csvFile
    self.nPausePosition = 0; # fangen von vorne an

    #setzen unsere Laufzeitprüfungen    
    self.isTestmode = False
    self.isFilesLoaded = False
    self.isFireShow = False

    #Laden von raspEasyFire
    self.nMaxChannel  = -1 # initial -1, beim csv laden wird es auf 0 gestezt
    
    #initialisren pygame und dessen player
    pygame.init()
    pygame.mixer.init(FREQ, BITSIZE, CHANNELS, BUFFER) #Werte kommen aus den Konstanten
    
    #intialisiere Display
    self.objLcd = RPi_I2C_driver.lcd()
    self.objLcd.backlight(1)
    
    #Laden uns den EventManager
    self.objEventManager = EventManager(self.objLcd)
  
  def loadFiles(self): 
    tmpFileError = None
    if not os.path.exists(self.csvFile):
      tmpFileError = "Csv File nicht gefunden"
    if not os.path.exists(self.sTestAudio):
      tmpFileError = "Testaudio nicht gefunden"
    if not os.path.exists(self.sFinalAudio):
      tmpFileError = "Finealaudio nicht gefunden"
    if tmpFileError is None:
      #Laden der Csv
      self.loadCsv(self.csvFile) 
      self.writeInfo("  ..waiting..  ", "  Lade Dateien  ", 3)
      self.isFilesLoaded = True
    else:
      self.writeError(tmpFileError);
      self.isFilesLoaded = False;

  def shutdown(self):
    #Display ausschalten
    self.objLcd.lcd_clear()
    self.objLcd.backlight(0)
    #Relai etc aus
    self.objRaspEasyFirePins.shutdown()
    #sind zuende, also machen wir alles fein zu
    pygame.quit()
  
  def writeError(self, sErrorMsg):
      if(hasattr(self.objLcd, 'tmpErrorMsg') and self.objLcd.tmpErrorMsg != sErrorMesg):
        self.objLcd.lcd_clear() # todo genrelle logik basteln

      self.objLcd.lcd_display_string("   ##ERROR##   ", 1)
      self.objLcd.lcd_display_string(sErrorMsg, 2)

  def writeInfo(self, sInfooTitle, sInfoMsg, nSleepTime):
      self.objLcd.lcd_clear()
      self.objLcd.lcd_display_string(sInfooTitle, 1)
      self.objLcd.lcd_display_string(sInfoMsg, 2)
      time.sleep(nSleepTime)

  def loadCsv(self, csvFile):
    self.nMaxChannel = -1
    with open('Silvester.csv', 'rb') as csvfile:
      objAktionReader = csv.reader(csvfile, delimiter=',', quotechar='"')
      for aktion in objAktionReader:
        if(self.nMaxChannel == -1):
          self.nMaxChannel = 0; #Headerline hat kein Kanal aber wir koennen die Datei lesen
          continue;

        # Haben eine weitere Zeile und müssen ein Wenig rechnen
        self.nMaxChannel = max(self.nMaxChannel, int(aktion[2]));
        tmpTime = aktion[3].split(":")
        tmpTime2 =tmpTime[1].split(".")
        objAktionTime = (((int(tmpTime[0]) * 60) + int(tmpTime2[0])) * 1000) + int(tmpTime2[1])
        self.objEventManager.setNewEvent(objAktionTime,int(aktion[2])) # setzen neue Vents zur Aktionlist       
  
  def loadi2cBus(self):
    if(self.objRaspEasyFirePins.isFireVoltageSw()):
      self.wirteError(" 12VOLT ist AN?")
      return False;
    else:
      #o.k Zündspannung ist aus, dann geben wir den Bus frei

      #5V relai ziehen!
      self.objRaspEasyFirePins.setBusVoltage(True)

      # MaximalChannel mit max Zündmodulen prüfen!
      #if(self.objEventManager.checkMaxChannel(self.nMaxChannel)):
        # setzen alle zündmodiule auf 0, daher wissen wir das wir sie erreichen!
        # war erfolgreich
      self.writeInfo("  ..waiting..  ","   Bus geladen! ", 3)
      return True        
      #else:
      #  return False


  def loadTestMode(self):
      if not (self.objRaspEasyFirePins.isFireVoltageSw()):
        #Keine Zündspannung also sind wir im Testmodus!
        if(self.loadi2cBus()):        
          #FireKnopf für Testaudio freigen (Leuchtet NICHT BLINKEN)
          self.objRaspEasyFirePins.setFireLed(True)
          self.isTestmode = True
          self.writeInfo("", "Sind im TESTMODE!!", 3)
        else:
          self.isTestmode = False
          self.writeError("   Busfehler!  ")
      else:
        self.writeError(" 12VOLT ist AN?")
        self.isTestmode = False

  def playAudioTest(self):
    pygame.mixer.music.load(self.sTestAudio)
    pygame.mixer.music.set_volume(1)
    pygame.mixer.music.play()
    while pygame.mixer.music.get_busy():
      if(self.objRaspEasyFirePins.isFireVoltageSw()):
        #haben Zündspannung gehen aus Testmodus
        pygame.mixer.music.stop();
      # CPU overkill bewaren!  
      time.sleep(0.030); 

    self.writeInfo("    Audio FERTIG  ","     ", 3)

  def runFireShow(self):
      print "Starting Firework"
      self.isFireShow = True
      self.objRaspEasyFirePins.setFireVoltage(True);
      pygame.mixer.music.load(self.sFinalAudio)
      pygame.mixer.music.play()

      #Laden uns das erste Event, rest passiert in der Schleife
      self.objEventManager.getNextEvent()
      while pygame.mixer.music.get_busy():
        if(self.objRaspEasyFirePins.isFireVoltageSw()):
          if(self.nPausePosition > 0):
            self.nPausePosition = 0
            pygame.mixer.music.unpause();
            self.objRaspEasyFirePins.setFireVoltage(True);
          
          self.objEventManager.TimeEvent(pygame.mixer.music.get_pos()); #check Audioposition im evenManager
          time.sleep(0.030); # CPU overkill bewaren!
        else:
          # NOTAUS, ZÜNDSPANNUNG WEG! 
          self.objRaspEasyFirePins.setFireVoltage(False);

          self.objRaspEasyFirePins.setFireLed(False)
          self.writeInfo("  12V to Resume  ","   !!NOTAUS!!   ", 1)
          self.nPausePosition = max(pygame.mixer.music.get_pos() - 5000, 0) # 5 Sekunfe vorher wirder starten
          pygame.mixer.music.pause();

      self.objRaspEasyFirePins.setFireVoltage(False); #jedenfal erstaml 12V weg!
      self.isFireShow = False

      #SAUBER DURCHGELAUFEN
      self.writeInfo("  FERTIG  ","   12V aus! ", 3)
        

  def setFireVoltage(self, isVoltage):
    print "setFireVoltage" +  str(isVoltage)
    #self.isFireVoltage = bool(isVoltage) # merken uns den zustand
    # in der FireShow regelt das selber für sich, ist dann Notaus!
    if(not self.isFireShow): 
      # 12 V Änderung, hier ist es ganz entscheidend ob Testmode oder nicht!
      if (not self.isTestmode and not isVoltage):
        print "Lade Testemode neu"
        #self.loadTestMode() # waren nicht im testmode und haben keine Zündspannung
      elif(self.isTestmode and not isVoltage):
        #sind im Testmode 
        self.objRaspEasyFirePins.setFireLed(True)
        self.writeInfo("","Sind im TESTMODE!!", 1)

      elif(self.isTestmode and isVoltage):
        #Zündspannung wurde angelegt, READY
        self.objRaspEasyFirePins.flashFireLed(True)
        self.writeInfo("      NOW  ","   !!READY!!  ", 1)

  def pushPower(self):
    if (not self.isFireShow):
      # in keiner Show also können wir uns ausmachen!
      self.writeInfo("      Shutdown","", 1)   
      #os.system("sudo shutdown -h now") 
    else:
      self.wirteError("  Show läuft!  ")
      self.isTestmode = False


  def pushFire(self):
    if(self.isTestmode and self.objRaspEasyFirePins.isFireVoltageSw()):
      # Testmodus geladen UND Zündspannung = READY!
      self.runFireShow();
    elif(self.isTestmode and not self.objRaspEasyFirePins.isFireVoltageSw()):
      # Testmode geladen und KEINE Zündspannung
      self.playAudioTest()

  def start(self):
    while True: 
      # Prüfen ob wir die Daten geladen haben
      if (self.isFilesLoaded):         
        #Files geladen - Bildschrim zeigt welcher Fehlt
        if (not self.isTestmode):
          # Prüfen den Bus und laden den Testmode
          self.loadTestMode()
        else:
          # haben alles und auf die Buttons zu warten
          print "Starting application logic"
          self.objRaspEasyFirePins.setWaitingLoop()
      else:
        # versuchen die Files zu laden
        self.loadFiles()
      
      # Sind wartezustand und laufen die checks erneut ab!
      time.sleep(.5) 

    




