#!/usr/bin/env python
# -*- coding: utf-8 -*-
# raspEasyFire, die RaspberryPi - Feuerwerskzündanlage
# Author, Jens Nachtigal - info@newan.de
# Weiternutzung und Veränderung bitte mit absprache

import RPi.GPIO as GPIO
import time
import threading
import Queue

CONST_PINONSW           = 7 # Pin für on/off  Taster
CONST_PINLEDONSW        = 11 # Pin für on/off  Taster- Led
CONST_PINFIRESW         = 13 # Pin für FIRE    Taster
CONST_PINLEDFIRESW      = 15 # Pin für FIRE    Taster- Led
CONST_PINFIREVOLTAGESW  = 16 # Pin für 12V     Switch
CONST_PINBUSVOLTAGE     = 24 # Pin für 5V      Relay-Switch
CONST_PINFIREVOLTAGE    = 26 # Pin für 12V     Relay-Switch


#Raspbarry Pin Klasse
class raspEasyFirePins:

  def __init__(self):
    # initial global parameter    
    self.toggleFire = False
    self.queue = Queue.Queue()

    #setup GPIO using Board numbering
    GPIO.setmode(GPIO.BOARD)
    
    # setzen unser Pins fürs Relai und setzen es direkt auf HIGH!:
    GPIO.setup(CONST_PINBUSVOLTAGE, GPIO.OUT)
    GPIO.setup(CONST_PINFIREVOLTAGE, GPIO.OUT)
    GPIO.output((CONST_PINBUSVOLTAGE, CONST_PINFIREVOLTAGE), GPIO.HIGH)

    # jetzt noch den Rest, nicht mehr ganz so riskant ;-)
    GPIO.setup(CONST_PINONSW,GPIO.IN, pull_up_down=GPIO.PUD_UP)#pullup
    GPIO.setup(CONST_PINLEDONSW, GPIO.OUT)
    GPIO.output(CONST_PINLEDONSW, GPIO.HIGH)
    GPIO.setup(CONST_PINFIRESW, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
    GPIO.setup(CONST_PINLEDFIRESW, GPIO.OUT)
    GPIO.output(CONST_PINLEDFIRESW, GPIO.LOW)
    GPIO.setup(CONST_PINFIREVOLTAGESW, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)

    # setzen Events für unsere 3 Buttons
    GPIO.add_event_detect(CONST_PINONSW, GPIO.RISING, callback=self.PowerSwitch, bouncetime=1000) #Nur an - aus ist uns egal
    GPIO.add_event_detect(CONST_PINFIRESW, GPIO.FALLING, callback=self.FireSwitch, bouncetime=1000)  #Nur an - aus ist uns egal
    GPIO.add_event_detect(CONST_PINFIREVOLTAGESW, GPIO.BOTH, callback=self.FireVoltageSwitch, bouncetime=200)  # Both damit wir an aus finden!


  def setRaspEasyFire(self, objRaspEasyFire):
    self.objRaspEasyFire = objRaspEasyFire

  #EventMethoden, die weitergeben werden zur Logik
  def PowerSwitch(self, channel):
    #self.objRaspEasyFire.pushPower()
    print "Todo beim Herunterfahren?"
    self.queue.put(channel)

  def FireSwitch(self, channel):
    self.queue.put(channel)  

  def FireVoltageSwitch(self, channel):
    if(GPIO.input(channel)):
      self.objRaspEasyFire.setFireVoltage(True)
      print "12V geändert auf True"
    else:
      self.objRaspEasyFire.setFireVoltage(False)
      print "12V geändert auf False"

  def isFireVoltageSw(self):
     return GPIO.input(CONST_PINFIREVOLTAGESW)

  # Relai Funktionen
  def setBusVoltage(self, bOn):
    if(bOn):
      GPIO.output(CONST_PINBUSVOLTAGE, GPIO.LOW) # Low is on!
    else:
      GPIO.output(CONST_PINBUSVOLTAGE, GPIO.HIGH)

  def setFireVoltage(self, bOn):
    if(bOn):
      GPIO.output(CONST_PINFIREVOLTAGE, GPIO.LOW) # Low is on!
    else:
      GPIO.output(CONST_PINFIREVOLTAGE, GPIO.HIGH)


  #Led funktionen
  def setPowerLed(self, bOn):
    if(bOn):
      GPIO.output((CONST_PINBUSVOLTAGE, CONST_PINLEDONSW), GPIO.HIGH)
    else:
      GPIO.output((CONST_PINBUSVOLTAGE, CONST_PINLEDONSW), GPIO.LOW)

  def setFireLed(self, bOn):
    #setzen festen wert KEIN Toggle:
    self.toggleFire = False
    time.sleep(0.5); # ! sekunde warten damit thread auch aufhört
    if(bOn):
      GPIO.output(CONST_PINLEDFIRESW, GPIO.HIGH)
    else:
      GPIO.output(CONST_PINLEDFIRESW, GPIO.LOW)
  
  def flashFireLed(self, bOn):
    print "Starting or Ending Flashing!"
    if(bOn):  
      self.toggleFire = True
      self.toggleFireLed()
    else:
      self.toggleFire = False

  def toggleFireLed(self):
    if(GPIO.input(CONST_PINLEDFIRESW)):
      GPIO.output(CONST_PINLEDFIRESW, GPIO.LOW)
    else:
      GPIO.output(CONST_PINLEDFIRESW, GPIO.HIGH)

    if self.toggleFire: #repeat it to toogle
      self.objFlashTimer = threading.Timer(0.5, self.toggleFireLed).start()

  # generell wartefunktio um auf eingaben zu warten
  def setWaitingLoop(self):
     while True:
        if not self.queue.empty():
          job = self.queue.get()
          if job == CONST_PINFIRESW:
            self.objRaspEasyFire.pushFire()
          elif job == CONST_PINONSW:
            self.objRaspEasyFire.pushPower()
        time.sleep(.2)

  def shutdown(self):
    # ganz wichtig den Bus und Zündung aus!
    self.toggleFire = False
    time.sleep(0.5) #Warten das thread zuende ist 
    self.setBusVoltage(False)
    self.setFireVoltage(False)

    GPIO.cleanup()# clean up GPIO on normal exit  


  

